/* StatServ routines

   OperStats - Copyright (c) 2000-2001 Darcy Grexton
   Contact: cygnus@uniserve.ca, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"
#include "../inc/ss-help.h"
#include "../inc/shared-help.h"

TLD *tldlist = NULL;
Boolean_T tldlookup = FALSE;

static void do_help (const char *source);
static void do_stats (const char *source);
static void do_map (const char *source);
static void do_servers (const char *source);
static void do_tld (const char *source);
static void do_tldmap (const char *source);
static void do_reset (const char *source);
static void do_flush (const char *source);
static void do_sendpings (const char *source);
static void sh_auth (const char *source);
static void sh_deauth (const char *source);
#ifdef RAWINJECT
static void sh_raw (const char *source);
static void sh_inject (const char *source);
#endif
static void sh_debug (const char *source);
static void sh_shutdown (const char *source);
static void sh_restart (const char *source);
static void sh_settings (const char *source);
static void sh_update (const char *source);
static void sh_rehash (const char *source);

static int showleafs (const char *source, char *server, char *prevpre,
		      char *arrowchar);
static int numleafs (char *server);

/* Main StatServ routine. */
void statserv (const char *source, char *buf)  
{
    char *cmd, *s;
    char orig[BUFSIZE];

    /* Make a copy of the original message. */ 
    strscpy (orig, buf, sizeof (orig));   

    cmd = strtok (buf, " ");

    if (!cmd)
	return;
    
    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_StatServ, source, "PING %s", s);
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
	ctcpreply (s_StatServ, source, "VERSION OperStats v%s", version);

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP",		H_NONE,		do_help},
	    {"STATS",		H_NONE,		do_stats},
	    {"TLD",		H_NONE,		do_tld},
	    {"TLDMAP",		H_NONE,		do_tldmap},
	    {"MAP",		H_IRCOP,	do_map},
	    {"SERVERS",		H_IRCOP,	do_servers},
	    {"AUTH",		H_IRCOP,	sh_auth},
	    {"DEAUTH",		H_SRA,		sh_deauth},
	    {"SHUTDOWN",	H_SRA,		sh_shutdown},
	    {"RESTART",		H_SRA,		sh_restart},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		sh_raw},
	    {"INJECT",		H_SRA,		sh_inject},
#endif
	    {"DEBUG",		H_SRA,		sh_debug},
	    {"UPDATE",		H_SRA,		sh_update},
	    {"REHASH",		H_SRA,		sh_rehash},
	    {"SETTINGS",	H_SRA,		sh_settings},
	    {"SENDPINGS",	H_SRA,		do_sendpings},
	    {"FLUSH",		H_SRA,		do_flush},
	    {"RESET",		H_SRA,		do_reset},
	    {NULL}
	};
            
	if ((command = get_hash (s_StatServ, source, strupper (cmd),
				 hash_table)))
	{
	    /* These are only available if OperServ is OFF. */
	    if ((command->process == sh_auth ||
		 command->process == sh_deauth ||
#ifdef RAWINJECT
		 command->process == sh_raw ||
		 command->process == sh_inject ||
#endif
		 command->process == sh_debug ||
		 command->process == sh_shutdown ||
		 command->process == sh_restart ||
		 command->process == sh_settings ||
		 command->process == sh_update ||
		 command->process == sh_rehash) &&
		 operserv_on == TRUE)
	    {
		notice (s_StatServ, source, ERR_UNKNOWN_COMMAND,
		    strupper (cmd), s_StatServ);
		return;
	    }

	    (*command->process) (source);

	    /* If OperServ is off, we want to snoop several commands. */
	    if ((command->process == sh_deauth ||
#ifdef RAWINJECT
		 command->process == sh_raw ||
		 command->process == sh_inject ||
#endif
		 command->process == sh_debug ||
		 command->process == sh_update ||
		 command->process == sh_rehash ||
		 command->process == sh_settings) &&
		 operserv_on == FALSE)
	    {
		log ("StatServ: <%s> %s", source, orig);
		snoop (s_StatServ, "<%s> %s", source, orig);
	    }
	    if (command->process == sh_auth)
	    {
		log ("StatServ: <%s> AUTH ...", source);
		snoop (s_StatServ, "<%s> AUTH ...", source);
	    }
	}
    }
}

/* Return a help message. */
static void do_help (const char *source) 
{
    char *cmd = strtok (NULL, " ");
            
    if (!cmd)
    {
	notice_list (s_StatServ, source, ss_help_start);
	notice_list (s_StatServ, source, ss_base_help);
	if (tldlookup == TRUE)
	    notice_list (s_StatServ, source, ss_help_tld);
	if (is_oper (source))
	    notice_list (s_StatServ, source, ss_oper_help);
	if (operserv_on == FALSE && is_oper (source))
	    notice_list (s_StatServ, source, shared_oper_help);
	if (operserv_on == FALSE && is_sra (source))
	    notice_list (s_StatServ, source, shared_sra_help);
	if (is_sra (source))
	    notice_list (s_StatServ, source, ss_sra_help);
	notice_list (s_StatServ, source, ss_help_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"STATS",		H_NONE,		stats_help},
	    {"TLD",		H_NONE,		tld_help},
	    {"TLDMAP",		H_NONE,		tldmap_help},
	    {"MAP",		H_IRCOP,	map_help},
	    {"SERVERS",		H_IRCOP,	servers_help},
	    {"AUTH",		H_IRCOP,	auth_help},
	    {"DEAUTH",		H_SRA,		deauth_help},
	    {"SHUTDOWN",	H_SRA,		shutdown_help},
	    {"RESTART",		H_SRA,		restart_help},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		raw_help},
	    {"INJECT",		H_SRA,		inject_help},
#endif
	    {"DEBUG",		H_SRA,		debug_help},
	    {"UPDATE",		H_SRA,		update_help},
	    {"REHASH",		H_SRA,		rehash_help},
	    {"SENDPINGS",	H_SRA,		sendpings_help},
	    {"SETTINGS",	H_SRA,		settings_help},
	    {"FLUSH",		H_SRA,		flush_help},
	    {"RESET",		H_SRA,		reset_help},
	    {NULL}
	};

	if ((command = get_help_hash (s_StatServ, source, strupper (cmd),
				      hash_table)))
	{
	    if (command->process == tld_help && tldlookup == FALSE)
	    {
		notice (s_StatServ, source, ERR_NO_HELP, "TLD");
		return;
	    }

	    /* These are only available if OperServ is OFF. */
	    if ((command->process == auth_help ||
		 command->process == deauth_help ||
		 command->process == shutdown_help ||
		 command->process == restart_help ||
#ifdef RAWINJECT
		 command->process == raw_help ||
		 command->process == inject_help ||
#endif
		 command->process == debug_help ||
		 command->process == update_help ||
		 command->process == rehash_help ||
		 command->process == settings_help) &&
		 operserv_on == TRUE)
	    {
		notice (s_StatServ, source, ERR_NO_HELP, strupper (cmd));
		return;
	    }

	    notice_list (s_StatServ, source, ss_help_start);
	    if (command->process == stats_help && is_oper (source))
		notice_list (s_StatServ, source, stats_help_oper);
	    else
		notice_list (s_StatServ, source, command->process);
	    notice_list (s_StatServ, source, ss_help_end);
	}
    }
}

/* Show some basic stats. */
static void do_stats (const char *source)
{
    long uptime;
    char *param = strtok (NULL, " ");

    if (!param || (param && !is_oper (source)))
    {
	uptime = (time (NULL) - me.since);

	notice (s_StatServ, source, SS_STATS_HEADER);
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_CURUSER, userstat);
	notice (s_StatServ, source, SS_STATS_MAXUSER, maxusercnt);
	if (is_oper (source))
	    notice (s_StatServ, source, SS_STATS_CURCHANOPER, countchans (0),
		countchans (1));
	else
	    notice (s_StatServ, source, SS_STATS_CURCHAN, countchans (0));
	notice (s_StatServ, source, SS_STATS_MAXCHAN, maxchancnt);
	notice (s_StatServ, source, SS_STATS_CURSERV, servstat);
	notice (s_StatServ, source, SS_STATS_MAXSERV, maxservcnt);
	notice (s_StatServ, source, SS_STATS_MAXHITS, maxhitcnt);

	/* If anyone can make this smaller, I'm listening ;) */
	if (speedsample > 1000000000)
	    notice (s_StatServ, source, SS_STATS_CURSPEEDGB, (float)
		speedsample / (float) 1000000000, (float) speedstat /
		(float) (speedstat > 1000000000 ? 1000000000 : speedstat >
		1000000 ? 1000000 : speedstat > 1000 ? 1000 : 1), speedstat >
		1000000000 ? "GB" : speedstat > 1000000 ? "MBs" : speedstat >
		1000 ? "KB" : "bytes");
	if (speedsample > 1000000)
	    notice (s_StatServ, source, SS_STATS_CURSPEEDMB, (float)
		speedsample / (float) 1000000, (float) speedstat /
		(float) (speedstat > 1000000000 ? 1000000000 : speedstat >
		1000000 ? 1000000 : speedstat > 1000 ? 1000 : 1), speedstat >
		1000000000 ? "GB" : speedstat > 1000000 ? "MB" : speedstat >
		1000 ? "KB" : "bytes");
	else if (speedsample > 1000)
	    notice (s_StatServ, source, SS_STATS_CURSPEEDKB, (float)
		speedsample / (float) 1000, (float) speedstat /
		(float) (speedstat > 1000000000 ? 1000000000 : speedstat >
		1000000 ? 1000000 : speedstat > 1000 ? 1000 : 1), speedstat >
		1000000000 ? "GB" : speedstat > 1000000 ? "MB" : speedstat >
		1000 ? "KB" : "bytes");
	else
	    notice (s_StatServ, source, SS_STATS_CURSPEEDBYTES, speedsample,
		(float) speedstat / (float) (speedstat > 1000000000 ?
		1000000000 : speedstat > 1000000 ? 1000000 : speedstat >
		1000 ? 1000 : 1), speedstat > 1000000000 ? "GB" : speedstat
		> 1000000 ? "MBs" : speedstat > 1000 ? "KB" : "bytes");
	if (is_oper (source))
	    notice (s_StatServ, source, SS_STATS_UPTIME, uptime/86400,
		uptime/86400 == 1 ? "" : "s", (uptime/3600) % 24,
		(uptime/60) %60, uptime % 60);
	notice (s_StatServ, source, SS_STATS_CURTIME, get_time (time (NULL),
	    1));
	notice (s_StatServ, source, SS_STATS_MAXUSERTIME, get_time
	    (maxusertime, 1));
	notice (s_StatServ, source, SS_STATS_MAXCHANTIME, get_time
	    (maxchantime, 1));
	notice (s_StatServ, source, SS_STATS_MAXSERVTIME, get_time
	    (maxservtime, 1));
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_DAILY, get_time (time (NULL),
	    0));
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_DAILY_HITS, dailyhitcnt);
	notice (s_StatServ, source, SS_STATS_DAILY_USERS, dailyuserstat,
	    get_time (dailyusertime, 2));
	notice (s_StatServ, source, SS_STATS_DAILY_CHANS, dailychanstat,
	    get_time (dailychantime, 2));
	notice (s_StatServ, source, SS_STATS_DAILY_SERVS, dailyservstat,
	    get_time (dailyservtime, 2));
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_FOOTER);
    }
    else
    {
	ServStat *serverstat;

	serverstat = findserv (param);

	if (!serverstat)
	{
	    notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	    return;
	}

	notice (s_StatServ, source, SS_SERVSTATS_HEADER, serverstat->name);
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_SERVSTATS_SINCE, get_time
	    (serverstat->statssince, 4), time_ago (serverstat->statssince));
	if (serverstat->splits)
	    notice (s_StatServ, source, SS_SERVSTATS_LASTSPLIT, get_time
		(serverstat->lastsplit, 4), time_ago (serverstat->lastsplit));
	notice (s_StatServ, source, SS_SERVSTATS_SPLITS, serverstat->splits,
	    serverstat->is_split ? " (Split)" : "");
	if (serverstat->lastquit)
	    notice (s_StatServ, source, SS_SERVSTATS_LASTQUIT,
		serverstat->lastquit);
	if (serverstat->is_split)
	{
	    if (serverstat->penalize)
		notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, time_ago
		    (time (NULL) - (serverstat->splittime + (time (NULL) -
		    serverstat->lastsplit))));
	    else
		notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, time_ago
		    (time (NULL) - serverstat->splittime));
	    notice (s_StatServ, source, SS_SERVSTATS_SPLITFOR, time_ago
		(serverstat->lastsplit));
	}
	else
	    notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, time_ago
		(time (NULL) - serverstat->splittime));
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_FOOTER);
    }
}

/* Show a listing of servers and their version replies. */
static void do_servers (const char *source)
{
    int i;
    ServStat *server;

    notice (s_StatServ, source, SS_SERVERS_HEADER);
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "%s (operstats-%s. %s%s%s%s) (0 splits)",
	me.name, version,
	(operserv_on == TRUE) ? "O" : "",
	(statserv_on == TRUE) ? "S" : "",
	(globalnoticer_on == TRUE) ? "G" : "",
	debuglevel ? "d" : "");   

    for (i = 0; i < 1024; i++)
    {
	for (server = statlist[i]; server; server = server->next)
	{
	    notice (s_StatServ, source, "%s (%s %s) (%d split%s)%s%s%s",
		server->name, server->version, server->flags, server->splits,
		server->splits == 1 ? "" : "s",
		server->is_split ? " (" : "",
		server->is_split ? "Split" : "",
		server->is_split ? ")" : "");
	}
    }
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, SS_SERVERS_FOOTER);
}

/* Display the country/group associated with a given TLD. */
static void do_tld (const char *source)
{
    TLD *tld;
    char *param = strtok (NULL, " ");

    if (tldlookup == FALSE)
    {
	notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, "TLD", s_StatServ);
	return;
    }

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "TLD TopLevelDomain");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "TLD");
	return;
    }

    tld = findtld (param);

    if (!tld)
    {
	notice (s_StatServ, source, SS_TLD_NOT_FOUND, param);
	return;
    }
    else
	notice (s_StatServ, source, SS_TLD_IS, tld->name, tld->desc);
}

/* Show either the user count/percentage for the given TLD, or
   show all active TLDs.
 */
static void do_tldmap (const char *source)
{
    TLD *tld;
    char *param = strtok (NULL, " ");

    if (tldlookup == FALSE)
    {
	notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, "TLDMAP", s_StatServ);
	return;
    }

    if (!param)
    {
	notice (s_StatServ, source, SS_TLDMAP_START);
	notice (s_StatServ, source, " ");

	for (tld = tldlist; tld; tld = tld->next)
	    if (tld->cnt)
		notice (s_StatServ, source, SS_TLDMAP, tld->desc,
		    " (", tld->name, ")", tld->cnt, tld->cnt == 1 ? "" : "s",
		    (float) (tld->cnt * 100) / usercnt);

	if (tldnores)
	    notice (s_StatServ, source, SS_TLDMAP, "Non Resolving", "", "", "",
		tldnores, tldnores == 1 ? "" : "s",
		(float) (tldnores * 100) / usercnt);

	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_TLDMAP_END);
    }
    else
    {
	tld = findtld (param);

	if (!tld)
	{
	    notice (s_StatServ, source, SS_TLD_NOT_FOUND, param);
	    return;
	}
	else
	    notice (s_StatServ, source, SS_TLDMAP, tld->desc, " (", tld->name,
		")", tld->cnt, tld->cnt == 1 ? "" : "s",
		(float) (tld->cnt * 100) / usercnt);
    }
}

/* Reset a given server's stats */
static void do_reset (const char *source)
{
    ServStat *server;
    char *param = strtok (NULL, " ");

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "RESET Server|ALL");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "RESET");
	return;
    }

    /* Reset a specific server */
    if (stricmp (param, "ALL"))
    {
	server = findserv (param);

	if (!server)
	{
	    notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	    return;
	}
	else
	{
	    server->statssince = time (NULL);
	    server->lastsplit = time (NULL);
	    server->splittime = 0;
	    if (server->is_split)
		server->splits = 1;
	    else
		server->splits = 0;
	    notice (s_StatServ, source, SS_SERVER_RESET, server->name);
	}
    }

    /* Reset ALL servers */
    else
    {
	int i;

	for (i = 0; i < 1024; i++)
	{
	    for (server = statlist[i]; server; server = server->next)
	    {
		server->statssince = time (NULL);
		server->lastsplit = time (NULL);
		server->splittime = 0;
		if (server->is_split)
		    server->splits = 1;
		else
		    server->splits = 0;
	    }
	}
	notice (s_StatServ, source, SS_SERVER_RESET, "all servers");
    }
}

/* FLUSH: Remove a server from the stats list (ie delinked) */
static void do_flush (const char *source)
{
    ServStat *server;
    char *param = strtok (NULL, " ");

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "FLUSH Server");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "FLUSH");
	return;
    }

    server = findserv (param);

    if (!server)
    {
	notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	return;
    }
    else
    {
	int i;

	/* Only flush servers that aren't online. */
	for (i = 0; i < servcnt; i++)
	{
	    if (!stricmp (servlist[i].server, param))
	    {
		notice (s_StatServ, source, SS_CANT_FLUSH);
		return;
	    }
	}
	free (server->version);
	free (server->flags);
	if (server->prev)
	    server->prev->next = server->next;
	else
	    statlist[SHASH(server->name)] = server->next;
	if (server->next)  
	    server->next->prev = server->prev;
	free (server);
	notice (s_StatServ, source, SS_SERVER_FLUSHED, param);
    }
}

/* Show a map of the network. */
static void do_map (const char *source)
{
    int i;
    char l[64];

    /* Find out which server we are connected to */
    for (i = 0; i < servcnt; ++i)
	if (servlist[i].hops == 1)
	    snprintf (l, sizeof (l), "%s", servlist[i].server);

    notice (s_StatServ, source, SS_MAP_HEADER);
    notice (s_StatServ, source, " ");
    showleafs (source, l, "", "");
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, SS_MAP_FOOTER);
}

/* Associated MAP function */
int showleafs (const char *source, char *server, char *prevpre,
	       char *arrowchar)
{
    User *user;
    int i = 0, j = 0, k = 0, x, num = 0, cnt, opcnt;
    char buf[BUFSIZE];
    char pre[BUFSIZE];
    char *nextpre;
    static int depth = 0;
 
    nextpre = malloc (BUFSIZE);
   
    snprintf (pre, sizeof (pre) + 1, prevpre);

    *buf = 0;

    for (i = 0; i < servcnt; i++)
    {
	cnt = 0;
	opcnt = 0;

	for (x = 0; x < 1024; x++)
	    for (user = userlist[x]; user; user = user->next)
		if (!strcmp (servlist[i].server, user->server))
		{
		    ++cnt;

		    if (is_oper (user->nick))
			++opcnt;
		}

	if (servlist[i].hops == 1 && match_wild_nocase (server,
							servlist[i].server))
	{
	    if (servlist[i].lag > 10)
		notice (s_StatServ, source,
		    "%s%s%s (%ds) (%d user%s (%d op%s), %3.2f%%)",
		    "\2", servlist[i].server, "\2", servlist[i].lag, cnt,
		    cnt == 1 ? "" : "s", opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	    else
		notice (s_StatServ, source,
		    "%s (%ds) (%d user%s (%d op%s), %3.2f%%)",
		    servlist[i].server, servlist[i].lag, cnt,
		    cnt == 1 ? "" : "s", opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	}
	else if (match_wild_nocase (server, servlist[i].server))
	{
	    strcat (buf, pre);
	    strcat (buf, arrowchar);
	    strcat (buf, "-");

	    if (servlist[i].lag > 10)
		strcat (buf, "\2");

	    strcat (buf, servlist[i].server);

	    if (servlist[i].lag > 10)
		notice (s_StatServ, source,
		    "%s%s (%ds) (%d user%s (%d op%s), %3.2f%%)",
		    buf, "\2", servlist[i].lag, cnt, cnt == 1 ? "" : "s",
		    opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	    else
		notice (s_StatServ, source,
		    "%s (%ds) (%d user%s (%d op%s), %3.2f%%)",
		    buf, servlist[i].lag, cnt, cnt == 1 ? "" : "s",
		    opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	}
    }

    j = numleafs (server);
    k = depth;

    for (i = 0; i < servcnt; i++)
    {
	depth = k;

	if (!stricmp (servlist[i].uplink, server))
	{
	    snprintf (nextpre, strlen (pre) + 1, pre);
	    if (depth++ > 0)
	    {
		if (arrowchar[0] == '`')
		    if (servlist[i].hops == 1)
			strcat (nextpre, "    ");
		    else
			strcat (nextpre, "  ");
		else
		    if (servlist[i].hops == 1)
			strcat (nextpre, "  | ");
		    else
			strcat (nextpre, "| ");
	    }
	    if (j > ++num)
		if (servlist[i].hops == 1)
		    showleafs (source, servlist[i].server, nextpre, "  |");
		else
		    showleafs (source, servlist[i].server, nextpre, "|");
	    else
		if (servlist[i].hops == 1)
		    showleafs (source, servlist[i].server, nextpre, "  `");
		else
		    showleafs (source, servlist[i].server, nextpre, "`");
	}
    }
    free (nextpre);
    depth = 0;
    return 1;
}

/* Associated MAP function */
int numleafs (char *server)
{
    int i, num = 0;
         
    for (i = 0; i < servcnt; ++i)
    {
	if (!stricmp (servlist[i].uplink, server))
	num++;
    }        
    return (num);
}

/* Load the StatServ DB from disk */
void load_ss_dbase (void)
{
    FILE *f = fopen (STATSERV_DB, "r");
    char *item, dBuf[2048], backup[2048];

    log ("Loading %s...", STATSERV_DB);

    if (!f)
    {
	log ("No %s found!", STATSERV_DB);

	/* Check for a db.save */
	strcpy (backup, STATSERV_DB);
	strcat (backup, ".save");
	if (rename (backup, STATSERV_DB) > 0)
	{
	    log ("Recovering %s.save", STATSERV_DB);
	    f = fopen (STATSERV_DB, "r");
	}
	else
	    return;
    }
    
    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");

	/* Ignore certain things.. */
	if (*item == '#' || *item == '\n' || *item == '\t' || *item == ' ' ||
	    *item == '\0' || *item == '\r')
	    continue;

	if (!strcmp (item, "ST"))
	{
	    maxusercnt	= atoi (strtok (NULL, " "));
	    maxusertime	= atol (strtok (NULL, " "));
	    maxchancnt	= atoi (strtok (NULL, " "));
	    maxchantime	= atol (strtok (NULL, " "));
	    maxservcnt	= atoi (strtok (NULL, " "));
	    maxservtime	= atol (strtok (NULL, " "));
	    maxhitcnt	= atoi (strtok (NULL, " "));
	    speedstat	= atoi (strtok (NULL, "\n"));
	}

	if (!strcmp (item, "SD"))
	{
	    dailyuserstat = atoi (strtok (NULL, " "));
	    dailyusertime = atol (strtok (NULL, " "));
	    dailychanstat = atoi (strtok (NULL, " "));
	    dailychantime = atol (strtok (NULL, " "));
	    dailyservstat = atoi (strtok (NULL, " "));
	    dailyservtime = atol (strtok (NULL, " "));
	    dailyhitcnt = atoi (strtok (NULL, "\n"));
	}

	if (!strcmp (item, "SS"))
	{
	    ServStat *serverstat, **list;

	    serverstat = scalloc (sizeof (ServStat), 1);

	    serverstat->name = sstrdup (strtok (NULL, " "));
	    serverstat->statssince = atol (strtok (NULL, " "));
	    serverstat->lastsplit = atol (strtok (NULL, " ")); 
	    serverstat->is_split = atoi (strtok (NULL, " "));
	    serverstat->splittime = atol (strtok (NULL, " "));
	    serverstat->splits = atoi (strtok (NULL, " "));
	    serverstat->lastquit = sstrdup (strtok (NULL, "\n"));
	    list = &statlist[SHASH(serverstat->name)];
	    serverstat->next = *list;
	    if (*list)
		(*list)->prev = serverstat;
	    *list = serverstat;
	}
    }
    fclose (f);
}

/* Write the stats db to disk. */
void save_ss_dbase ()
{
    FILE *f;
    ServStat *server;
    int i;
    char backup[2048];

    if (logupdates_on == TRUE)
	log ("Saving %s...", STATSERV_DB);

    strcpy (backup, STATSERV_DB);  
    strcat (backup, ".save");
    remove (backup);
    if (rename (STATSERV_DB, backup) < 0)
	log ("Can't back up %s!", STATSERV_DB);
    f = fopen (STATSERV_DB, "w");
    
    if (!f)
    {
	log ("Can't write to %s!", STATSERV_DB);
	if (rename (backup, STATSERV_DB) < 0)
	    fatal ("Can't restore backup of %s!", STATSERV_DB);
	return;
    }
        
    /* Print an identifying line at the top of the DB. */
    fprintf (f, "OperStats v%s %s\n", version, STATSERV_DB);

    /* Various Statistics */
    fprintf (f, "ST %d %lu %d %lu %d %lu %d %d\n", maxusercnt, maxusertime,
	maxchancnt, maxchantime, maxservcnt, maxservtime, maxhitcnt,
	speedstat);

    /* Stats for the current day */
    fprintf (f, "SD %d %lu %d %lu %d %lu %d\n", dailyuserstat, dailyusertime,
	dailychanstat, dailychantime, dailyservstat, dailyservtime,
	dailyhitcnt);

    /* Per-Server stats */
    for (i = 0; i < 1024; i++)
    {
	for (server = statlist[i]; server; server = server->next)
	{
	    fprintf (f, "SS %s %lu %lu %d %d %d %s\n", server->name,
		server->statssince, server->lastsplit, server->is_split,
		server->splittime, server->splits, server->lastquit ?
		server->lastquit : "N/A");
	}
    }

    fclose (f);
    remove (backup);
}

/* Add a new TLD into the tld list. */
static TLD * new_tld (const char *name, const char *desc)
{
    TLD *tld;

    if ((tld = findtld (name))) /* Don't want dupes, rehash */
    {
	tld->desc = sstrdup (desc);
	return 0;
    }

    tld = scalloc (sizeof (TLD), 1);

    tld->name = sstrdup (name);
    tld->desc = sstrdup (desc);

    tld->next = tldlist;

    if (tld->next)
	tld->next->prev = tld;

    tldlist = tld;

    return tld;
}

/* Load the TLD listings */
void load_tlds (void)
{
    FILE *f = fopen (TLDFILE, "r");
    char *s, dBuf[BUFSIZE], *name, *desc;

    log ("Loading %s...", TLDFILE);

    if (!f)
    {
	tldlookup = FALSE;
	return;
    }

    while (fgets (dBuf, 2047, f))
    {
	s = strtok (dBuf, " ");
	strip (s);

	if (s)
	    name = s;
	else
	    name = "";

	s = strtok (NULL, "");
	strip (s);

	if (s)
	    desc = s;
	else
	    desc = "";

	if (!desc)
	    log ("Found TLD %s but no description!", name);
	else
	    new_tld (name, desc);
    }
    fclose (f);
    tldlookup = TRUE;
}

/* Ping all servers. */
void do_sendpings (const char *source)
{
    int i;

    for (i = 0; i < servcnt; ++i)
    {
	send_cmd (me.name, "%s %s %s", me.token ? "8" : "PING",
	    me.name, servlist[i].server);
	servlist[i].lastping = time (NULL);
    }

    notice (s_StatServ, source, SS_PINGED_ALL);
}

/* Reset the daily statistics to current values. */
void reset_daily_stats (void)
{
    dailyuserstat = userstat;
    dailyusertime = time (NULL);
    dailychanstat = chanstat;
    dailychantime = time (NULL);
    dailyservstat = servstat;
    dailyservtime = time (NULL);
}

/* Shared functions: These call shared functions with the appropriate
   information.
 */
static void sh_auth (const char *source)
{
    do_auth (s_StatServ, source);
}

static void sh_deauth (const char *source)
{
    do_deauth (s_StatServ, source);
}

#ifdef RAWINJECT
static void sh_raw (const char *source)
{
    do_raw (s_StatServ, source);
}

static void sh_inject (const char *source)
{
    do_inject (s_StatServ, source);
}
#endif

static void sh_debug (const char *source)
{
    do_debug (s_StatServ, source);
}           

static void sh_shutdown (const char *source)
{
    do_shutdown (s_StatServ, source);
}

static void sh_restart (const char *source)
{
    do_restart (s_StatServ, source);
}

static void sh_settings (const char *source)
{
    do_settings (s_StatServ, source);
}

static void sh_update (const char *source)
{
    do_update (s_StatServ, source);
}

static void sh_rehash (const char *source)
{
    do_rehash (s_StatServ, source);
}
